# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Compute spatially-resolved operative temperature and PMV thermal comfort from
a Honeybee model. This recipe can also (optionally) compute Standard Effective
Temperature (SET).
_
This recipe uses EnergyPlus to obtain longwave radiant temperatures and indoor air
temperatures. The outdoor air temperature and air speed are taken directly from
the EPW. A Radiance-based, enhanced 2-phase method is used for shortwave MRT
calculations, which uses an accurate direct sun calculation with precise solar
positions.
_
The energy properties of the model geometry are what determine the outcome of the
simulation, though the model's Radiance sensor grids are what determine where
the comfort mapping occurs.
-

    Args:
        _model: A Honeybee Model for which PMV comfort will be mapped. Note that
            this model should have radiance grids assigned to it in order
            to produce meaningful results.
        _epw: Path to an EPW weather file to be used for the comfort map simulation.
        _ddy: Path to a DDY file with design days to be used for the initial sizing
            calculation of the energy simulation.
        north_: A number between -360 and 360 for the counterclockwise difference
            between the North and the positive Y-axis in degrees. This can
            also be Vector for the direction to North. (Default: 0).
        run_period_: An AnalysisPeriod to set the start and end dates of the simulation.
            If None, the simulation will be annual.
        sensor_count_: Integer for the maximum number of sensor grid points per
            parallel execution. (Default: 200).
        write_set_map_: A boolean to note whether the output temperature CSV should
            record Operative Temperature or Standard Effective Temperature (SET).
            SET is relatively intense to compute and so only recording Operative
            Temperature can greatly reduce run time, particularly when air speeds
            are low. However, SET accounts for all 6 PMV model inputs and so is a
            more representative "feels-like" temperature for the PMV model.
        _air_speed_: A single number for air speed in m/s or an hourly data collection
            of air speeds that align with the input run_period_. This will be
            used for all indoor comfort evaluation. Note that the EPW wind speed
            will be used for any outdoor sensors. (Default: 0.1).
        _met_rate_: A single number for metabolic rate in met or an hourly data collection
            of met rates that align with the run_period_. (Default: 1.1, for
            seated, typing).
        _clo_value_: A single number for clothing level in clo or an hourly data collection
            of clothing levels that align with the run_period_. (Default: 0.7,
            for pants and a long sleeve shirt).
        comfort_par_: Optional comfort parameters from the "LB PMV Comfort Parameters"
            component to specify the criteria under which conditions are
            considered acceptable/comfortable. The default will assume a
            PPD threshold of 10% and no absolute humidity constraints.
        solar_body_par_: Optional solar body parameters from the "LB Solar Body Parameters"
            object to specify the properties of the human geometry assumed in the
            shortwave MRT calculation. The default assumes average skin/clothing
            absorptivity and a human subject always has their back to the sun
            at a 45-degree angle (SHARP = 135).
        radiance_par_: Text for the radiance parameters to be used for ray
            tracing. (Default: -ab 2 -ad 5000 -lw 2e-05).
        run_settings_: Settings from the "HB Recipe Settings" component that specify
            how the recipe should be run. This can also be a text string of
            recipe settings.
        _run: Set to True to run the recipe and get results.

    Returns:
        report: Reports, errors, warnings, etc.
        temperature: A folder containing CSV maps of Operative Temperature for each sensor
            grid at each time step of the analysis. Alternatively, if the
            write_set_map_ option is used, the CSV maps here will contain
            Standard Effective Temperature (SET). This can be connected to the
            "HB Read Thermal Matrix" component to parse detailed results into
            Grasshopper. Values are in Celsius.
        condition: A folder containing CSV maps of comfort conditions for each sensor
            grid at each time step of the analysis. This can be connected to the
            "HB Read Thermal Matrix" component to parse detailed results into
            Grasshopper. Values are as follows.
            _
            .    -1 = unacceptably cold conditions
            .     0 = neutral (comfortable) conditions
            .    +1 = unacceptably hot conditions
        pmv: A folder containing CSV maps of the Predicted Mean Vote (PMV) for each
            sensor grid at each time step of the analysis. This can be connected
            to the "HB Read Thermal Matrix" component to parse detailed results
            into Grasshopper. This can be used to understand not just whether
            conditions are acceptable but how uncomfortably hot or cold they are.
        TCP: Lists of values between 0 and 100 for the Thermal Comfort Percent (TCP).
            These can be plugged into the "LB Spatial Heatmap" component along
            with meshes of the sensor grids to visualize spatial thermal comfort.
            TCP is the percentage of occupied time where thermal conditions are
            acceptable/comfortable. Occupied hours are determined from the
            occupancy schedules of each room (any time where the occupancy
            schedule is >= 0.1 will be considered occupied). Outdoor sensors
            are considered occupied at all times. More custom TCP studies can
            be done by post-processing the condition results.
        HSP: Lists of values between 0 and 100 for the Heat Sensation Percent (HSP).
            These can be plugged into the "LB Spatial Heatmap" component along with
            meshes of the sensor grids to visualize uncomfortably hot locations.
            HSP is the percentage of occupied time where thermal conditions are
            hotter than what is considered acceptable/comfortable. Occupied hours
            are determined from the occupancy schedules of each room (any time
            where the occupancy schedule is >= 0.1 will be considered occupied).
            Outdoor sensors are considered occupied at all times. More custom HSP
            studies can be done by post-processing the condition results.
        CSP: Lists of values between 0 and 100 for the Cold Sensation Percent (CSP).
            These can be plugged into the "LB Spatial Heatmap" component along with
            meshes of the sensor grids to visualize uncomfortably cold locations.
            CSP is the percentage of occupied time where thermal conditions are
            colder than what is considered acceptable/comfortable. Occupied hours
            are determined from the occupancy schedules of each room (any time
            where the occupancy schedule is >= 0.1 will be considered occupied).
            Outdoor sensors are considered occupied at all times. More custom CSP
            studies can be done by post-processing the condition results.
"""

ghenv.Component.Name = 'HB PMV Comfort Map'
ghenv.Component.NickName = 'PMVMap'
ghenv.Component.Message = '1.2.1'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '7 :: Thermal Map'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from lbt_recipes.recipe import Recipe
except ImportError as e:
    raise ImportError('\nFailed to import lbt_recipes:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, recipe_result
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _run:
    # create the recipe and set the input arguments
    recipe = Recipe('pmv_comfort_map')
    recipe.input_value_by_name('model', _model)
    recipe.input_value_by_name('epw', _epw)
    recipe.input_value_by_name('ddy', _ddy)
    recipe.input_value_by_name('north', north_)
    recipe.input_value_by_name('run-period', run_period_)
    recipe.input_value_by_name('sensor-count', sensor_count_)
    recipe.input_value_by_name('write-set-map', write_set_map_)
    recipe.input_value_by_name('air-speed', _air_speed_)
    recipe.input_value_by_name('met-rate', _met_rate_)
    recipe.input_value_by_name('clo-value', _clo_value_)
    recipe.input_value_by_name('comfort-parameters', comfort_par_)
    recipe.input_value_by_name('solarcal-parameters', solar_body_par_)
    recipe.input_value_by_name('radiance-parameters', radiance_par_)

    # run the recipe
    project_folder = recipe.run(
        run_settings_, radiance_check=True, openstudio_check=True)

    # load the results
    temperature = recipe_result(recipe.output_value_by_name('temperature', project_folder))
    condition = recipe_result(recipe.output_value_by_name('condition', project_folder))
    pmv = recipe_result(recipe.output_value_by_name('pmv', project_folder))
    TCP = recipe_result(recipe.output_value_by_name('tcp', project_folder))
    HSP = recipe_result(recipe.output_value_by_name('hsp', project_folder))
    CSP = recipe_result(recipe.output_value_by_name('csp', project_folder))



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Get information about end use intensity from an EnergyPlus SQL file.
-

    Args:
        _sql: The file path of the SQL result file that has been generated from
            an energy simulation. This can also be a list of EnergyPlus files
            in which case, EUI will be computed across all files.
        ip_: Boolean to note whether the EUI should be in SI (kWh/m2) or IP
            (kBtu/ft2) units. (Default: False).

    Returns:
        eui: The total end use intensity result from the simulation. Specifically,
            this is the sum of all electricity, fuel, district heating/cooling,
            etc. divided by the gross floor area (including both conditioned
            and unconditioned spaces). The value will be in kWh/m2 if ip_
            is False or None and kBtu/ft2 if True.
        eui_end_use: The end use intensity result from the simulation, broken down by each
            end use. These values coorespond to the end_uses output below. Values
            will be in kWh/m2 if ip_ is False or None and kBtu/ft2 if True.
        end_uses: A list of text for each of the end uses in the simulation (Heating,
            Cooling, etc.). Thes outputs coorespond to the eui_end_use
            output above.
        gross_floor: The total gross floor area of the energy model. This can be used
            to compute the total energy use from the intensity values above or
            it can be used to help with other result post-processing. The value
            will be in m2 if ip_ is False or None and ft2 if True.
"""

ghenv.Component.Name = 'HB End Use Intensity'
ghenv.Component.NickName = 'EUI'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '6 :: Result'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

from collections import OrderedDict

try:
    from ladybug.sql import SQLiteResult
    from ladybug.datatype.area import Area
    from ladybug.datatype.energyintensity import EnergyIntensity
    from ladybug.datatype.energy import Energy
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # set initial values that will be computed based on results
    total_floor_area, total_energy = 0, 0
    all_uses = \
        ('heating', 'cooling', 'interior_lighting', 'exterior_lighting',
         'interior_equipment', 'exterior_equipment', 'fans', 'pumps',
         'heat_rejection', 'humidification', 'heat_recovery', 'water_systems',
          'refrigeration', 'generators')
    end_uses = OrderedDict()
    for use in all_uses:
        end_uses[use] = 0

    # loop through the sql files in the directory and add the energy use
    for result_file in _sql:
        # parse the SQL file
        sql_obj = SQLiteResult(result_file)
        # get the total floor area of the model
        area_dict = sql_obj.tabular_data_by_name('Building Area')
        areas = tuple(area_dict.values())
        total_floor_area += areas[0][0]
        # get the energy use
        eui_dict = sql_obj.tabular_data_by_name('End Uses')
        euis = tuple(eui_dict.values())
        total_energy += sum([val for val in euis[-2][:12]])
        end_uses['heating'] += sum([val for val in euis[0][:12]])
        end_uses['cooling'] += sum([val for val in euis[1][:12]])
        end_uses['interior_lighting'] += sum([val for val in euis[2][:12]])
        end_uses['exterior_lighting'] += sum([val for val in euis[3][:12]])
        end_uses['interior_equipment'] += sum([val for val in euis[4][:12]])
        end_uses['exterior_equipment'] += sum([val for val in euis[5][:12]])
        end_uses['fans'] += sum([val for val in euis[6][:12]])
        end_uses['pumps'] += sum([val for val in euis[7][:12]])
        end_uses['heat_rejection'] += sum([val for val in euis[8][:12]])
        end_uses['humidification'] += sum([val for val in euis[9][:12]])
        end_uses['heat_recovery'] += sum([val for val in euis[10][:12]])
        end_uses['water_systems'] += sum([val for val in euis[11][:12]])
        end_uses['refrigeration'] += sum([val for val in euis[12][:12]])
        end_uses['generators'] += sum([val for val in euis[13][:12]])

    # assemble all of the results into a final dictionary
    eui = round(total_energy / total_floor_area, 3)
    gross_floor = round(total_floor_area, 3)
    end_use_pairs = [(key, round(val / total_floor_area, 3))
                     for key, val in end_uses.items() if val != 0]
    eui_end_use = [use[1] for use in end_use_pairs]
    end_uses = [use[0].replace('_', ' ').title() for use in end_use_pairs]

    # convert data to IP if requested
    if ip_:
        eui_typ, a_typ, e_typ = EnergyIntensity(), Area(), Energy()
        eui = round(eui_typ.to_ip([eui], 'kWh/m2')[0][0], 3)
        gross_floor = round(a_typ.to_ip([gross_floor], 'm2')[0][0], 3)
        eui_end_use = [round(eui_typ.to_ip([val], 'kWh/m2')[0][0], 3)
                       for val in eui_end_use]



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Create a list of exterior constructions that can be used to edit or create a
ConstructionSet object.
-

    Args:
        _exterior_wall_: A construction object for exterior walls (or text for
            the identifier of the construction within the library).
        _exterior_roof_: A construction object for exterior roofs (or text for
            the identifier of the construction within the library).
        _exposed_floor_: A construction object for exposed floors (or text for
            the identifier of the construction within the library).
    
    Returns:
        exterior_set: A list of exterior constructions that can be used to edit
            or create a ConstructionSet object.
"""

ghenv.Component.Name = "HB Exterior Construction Subset"
ghenv.Component.NickName = 'ExteriorSubset'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '0 :: Basic Properties'
ghenv.Component.AdditionalHelpFromDocStrings = "4"

try:  # import honeybee_energy dependencies
    from honeybee_energy.construction.opaque import OpaqueConstruction
    from honeybee_energy.lib.constructions import opaque_construction_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))


def opaque_constr(construction, input_name):
    """Get an OpaqueConstrucion from the library if it's a string."""
    if isinstance(construction, str):
        return opaque_construction_by_identifier(construction)
    else:
        assert isinstance(construction, OpaqueConstruction), \
            'Expected OpaqueConstruction for {}. Got {}'.format(
                input_name, type(construction))
    return construction


# go through each input construction
if _exterior_wall_ is not None:
    _exterior_wall_ = opaque_constr(_exterior_wall_, '_exterior_wall_')
if _exterior_roof_ is not None:
    _exterior_roof_ = opaque_constr(_exterior_roof_, '_exterior_roof_')
if _exposed_floor_ is not None:
    _exposed_floor_ = opaque_constr(_exposed_floor_, '_exposed_floor_')


# return the final list from the component
exterior_set = [_exterior_wall_, _exterior_roof_, _exposed_floor_]
